// FileManager.cpp

#include "stdafx.h"
#include "afx.h"
#include "atlstr.h"

#include "FileManager.h"


#pragma warning ( disable : 4018 )

//------------------------------------------------------------------------------

#undef MAX_PATH
#define MAX_PATH 65536

//------------------------------------------------------------------------------

FileManager& getFileManager()
{
	return FileManager::getInstance();
}

//------------------------------------------------------------------------------
// AppendSlash
//------------------------------------------------------------------------------

bool AppendSlash ( std::string & path )
{
	if ( path.empty() || path[path.size()-1] != '\\' )
	{
		path.append ( "\\" );
		return true;
	}
	return false;
}

//------------------------------------------------------------------------------
// ConcatPathFile
//
// conceit path with relative path or file
//------------------------------------------------------------------------------

bool ConcatPathFile ( const char * path, const char * file, std::string & path_file )
{
	if ( !path || !file )
	{
		ASSERT(0);
		path_file = std::string();
		return false;
	}
	
	std::string tmp_path ( path );
	AppendSlash ( tmp_path );
	if ( file[0] != '\\' )
		tmp_path.append ( file );
	else
		tmp_path.append ( & file[1] );
	path_file = tmp_path;
	return true;
}

//------------------------------------------------------------------------------
// GetFileFromPath
//------------------------------------------------------------------------------

bool GetFileFromPath ( const char * path, std::string & file )
{
	if ( !path )
	{
		ASSERT(0);
		file = std::string();
		return false;
	}
	
	int i_end = (int)strlen ( path ) - 1;
	int i;
	for ( i = i_end; i >=0 ; --i )
	{
		if ( path[i] == '\\' )
		{
			++i;
			break;
		}
	}
	if ( i < 0 )
		i = 0;
	file.assign ( & path[i] );
	return true;
}

//------------------------------------------------------------------------------
// GetDirectoryFromPath
//------------------------------------------------------------------------------

bool GetDirectoryFromPath ( const char * path, std::string & directory )
{
	if ( !path )
	{
		ASSERT(0);
		directory = std::string();
		return false;
	}
	
	int i_end = (int)strlen ( path ) - 1;
	int i;
	for ( i = i_end; i >=0 ; --i )
	{
		if ( path[i] == '\\' )
		{
			break;
		}
	}

	if ( i >= 0 )
		directory.assign ( path, i );
	return true;
}

//------------------------------------------------------------------------------

bool GetRelativePath
(
	const char * file_path,
	const char * base_path,
	std::string & relative_path
)
{
	bool bCompare = true;
	bool bDoubleQuotes = false;

	while ( *file_path && ( *file_path == *base_path ) )
	{
		bCompare = false;
		if ( bDoubleQuotes )
		{
			if ( *file_path == '\"' )
			{
				bDoubleQuotes = false;
			}
		}
		else
		{
			if ( *file_path == '\"' )
			{
				bDoubleQuotes = true;
			}
			else
			{
				if ( *file_path == ':' || *file_path == '\\' )
					bCompare = true;
			}
		}
		++file_path;
		++base_path;
	}
	if ( !bCompare )
	{
		if ( *file_path == '\\' && *base_path == '\0' )
		{
			++file_path;
			bCompare = true;
		}
	}
	relative_path.assign ( file_path );
	return bCompare;
}

//------------------------------------------------------------------------------

bool GetFullPathFromRelative
(
	const char * relative_path,
	std::string & full_path
)
{
	char buffer[MAX_PATH];
	char * file_in_buffer;
	DWORD dword = ::GetFullPathName ( relative_path, MAX_PATH, buffer, & file_in_buffer );

	if ( dword )
		full_path.assign ( buffer );
	else
		full_path.clear();

	return dword != 0;
}

//------------------------------------------------------------------------------

bool EvalPath ( const char * source_path, const char * input_base_path, const char * output_base_path, std::string & target_path )
{
	std::string relativePath;
	
	if ( !GetRelativePath ( source_path, input_base_path, relativePath ) )
		return false;

	if ( !ConcatPathFile ( output_base_path, relativePath.c_str(), target_path ) )
		return false;

	return true;
}

//------------------------------------------------------------------------------
// File
//------------------------------------------------------------------------------

File::File ( const char * file )
	: handle_(INVALID_HANDLE_VALUE)
{
	char buffer[MAX_PATH];
	char * file_in_buffer;
	
	DWORD dword = ::GetFullPathName ( file, MAX_PATH, buffer, & file_in_buffer );
	
	full_path_.assign ( buffer );
	::GetFileFromPath ( full_path_.c_str(), file_ );
}

//------------------------------------------------------------------------------

File::~File()
{
	close();
}

//------------------------------------------------------------------------------

bool File::open ( E_access_mode open_mode )
{
	switch ( open_mode )
	{
		case ac_read:
			{
				handle_ = ::CreateFile
				(
					full_path_.c_str(),
					GENERIC_READ,
					FILE_SHARE_READ,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					NULL
				);
			}
			break;

		case ac_write:
			{
				handle_ = ::CreateFile
				(
					full_path_.c_str(),
					GENERIC_WRITE,
					0,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					NULL
				);
			}
			break;

		case ac_read_write:
			{
				handle_ = ::CreateFile
				(
					full_path_.c_str(),
					GENERIC_READ | GENERIC_WRITE,
					0,
					NULL,
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,
					NULL
				);
			}
			break;

		case ac_create:
			{
				handle_ = ::CreateFile
				(
					full_path_.c_str(),
					GENERIC_WRITE,
					FILE_SHARE_WRITE,
					NULL,
					CREATE_ALWAYS,
					FILE_ATTRIBUTE_NORMAL,
					NULL
				);
			}
			break;

		default:
			return false;
	}

	if ( handle_ == INVALID_HANDLE_VALUE )
		return false;
		//throw FileError ( FileError::cannot_open_file, full_path_ );

	return true;
}

//------------------------------------------------------------------------------

void File::close()
{
	if ( handle_ != INVALID_HANDLE_VALUE )
	{
		::CloseHandle ( handle_ );
		handle_ = INVALID_HANDLE_VALUE;
	}
}

//------------------------------------------------------------------------------

bool File::read ( void * buffer, size_t bytes, size_t & readed )
{
	DWORD bytes_read;
	
	BOOL b = ::ReadFile
	(
		handle_,
		buffer,
		(DWORD)bytes,
		& bytes_read,
		NULL
	);

	readed = bytes_read;
	return b != FALSE;
}

//------------------------------------------------------------------------------

bool File::read ( void * buffer, size_t bytes )
{
	DWORD bytes_read;

	BOOL b = ::ReadFile
	(
		handle_,
		buffer,
		(DWORD)bytes,
		& bytes_read,
		NULL
	);

	return b != FALSE && bytes_read == bytes;
}

//------------------------------------------------------------------------------

bool File::write ( const void * buffer, size_t bytes, size_t & written )
{
	DWORD bytes_write;
	
	BOOL b = ::WriteFile
	(
		handle_,
		buffer,
		(DWORD)bytes,
		& bytes_write,
		NULL
	);

	written = bytes_write;
	return b != FALSE;
}

//------------------------------------------------------------------------------

bool File::write ( const void * buffer, size_t bytes )
{
	DWORD bytes_write;
	
	BOOL b = ::WriteFile
	(
		handle_,
		buffer,
		(DWORD)bytes,
		& bytes_write,
		NULL
	);

	return b != FALSE && bytes_write == bytes;
}

//------------------------------------------------------------------------------

bool File::seekToEndOfFile()
{
	DWORD dword =	SetFilePointer (
		handle_,
		0,
		0,
		FILE_END
	);

	return dword != INVALID_SET_FILE_POINTER;
}

//------------------------------------------------------------------------------

bool File::getLength ( size_t & length )
{
	DWORD lpFileSizeHigh;

	length = ::GetFileSize ( handle_, & lpFileSizeHigh );

	return length != INVALID_FILE_SIZE;
}

//------------------------------------------------------------------------------

bool File::getName ( std::string & file_name )
{
	file_name = file_;
	return true;
}

//------------------------------------------------------------------------------

bool File::getFullPath ( std::string & full_path )
{
	full_path = full_path_;
	return true;
}

//------------------------------------------------------------------------------

bool File::getDirectory ( std::string & directory )
{
	return GetDirectoryFromPath ( full_path_.c_str(), directory );
}

//------------------------------------------------------------------------------

bool File::fileExistAndNotDirectory() const
{
	DWORD dword = GetFileAttributes ( full_path_.c_str() );
	return ( dword != INVALID_FILE_ATTRIBUTES
		&& ( ( dword & FILE_ATTRIBUTE_DIRECTORY ) == 0 ) );
}

//------------------------------------------------------------------------------

//void setError ( PFile file, const std::string & suffix )
//{
//	std::string file_name;
//	if ( !file->getName ( file_name ) )
//	{
//		CString s;
//		s.Format ( "file %p", file );
//		file_name.assign ( (LPCTSTR)s );
//	}
//	setError ( file_name + suffix );
//}

//------------------------------------------------------------------------------
// FileIterator
//------------------------------------------------------------------------------

FileIterator::FileIterator()
{}

FileIterator::~FileIterator()
{}

//------------------------------------------------------------------------------

namespace FileNamespace
{

//------------------------------------------------------------------------------
// FileIteratorWithPred
//------------------------------------------------------------------------------

Win32FileIteratorWithPred::Win32FileIteratorWithPred
(
	const char * file_mask,
	file_pred_fun fun
)
: file_mask_(file_mask),
	fun_(fun),
	handle_(INVALID_HANDLE_VALUE),
	bEnd_(false)
{}

Win32FileIteratorWithPred::~Win32FileIteratorWithPred()
{
	if ( handle_ != INVALID_HANDLE_VALUE )
	{
		::FindClose ( handle_ );
		handle_ = INVALID_HANDLE_VALUE;
	}
}

bool Win32FileIteratorWithPred::next ( WIN32_FIND_DATA & findFileData )
{
	if ( bEnd_ )
		return false;

next:
	if ( handle_ == INVALID_HANDLE_VALUE )
	{
    handle_ = ::FindFirstFile ( file_mask_.c_str(), & findFileData );

		if ( handle_ == INVALID_HANDLE_VALUE )
		{
			bEnd_ = true;
			return false;
		}
	}
	else
	{
		BOOL bFileFound = ::FindNextFile ( handle_, & findFileData );
		if ( !bFileFound )
		{
			::FindClose ( handle_ );
			handle_ = INVALID_HANDLE_VALUE;
			bEnd_ = true;
			return false;
		}
	}

	if ( !fun_ ( findFileData ) )
		goto next;

	return true;
}

//------------------------------------------------------------------------------

bool isFile ( const WIN32_FIND_DATA & findFileData )
{
	return ( ( findFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != FILE_ATTRIBUTE_DIRECTORY );
}

//------------------------------------------------------------------------------

bool isDirectory ( const WIN32_FIND_DATA & findFileData )
{
	return ( ( findFileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == FILE_ATTRIBUTE_DIRECTORY );
}

//------------------------------------------------------------------------------

bool isSubDirectory ( const WIN32_FIND_DATA & findFileData )
{
	if ( isDirectory ( findFileData ) )
	{
		std::string s ( findFileData.cFileName );
		return ( s != "." && s != ".." );
	}
	else
		return false;
}

//------------------------------------------------------------------------------

bool getCurrentDirectory ( std::string & s )
{
	char current_directory[MAX_PATH];
	char full_path[MAX_PATH];
	DWORD cur_dir_size = GetCurrentDirectory ( MAX_PATH, current_directory );
	if ( !cur_dir_size )
	{
		ASSERT(0);
		return false;
	}

	LPTSTR notused;
	DWORD full_path_size = ::GetFullPathName ( current_directory, MAX_PATH, full_path, & notused );
	if ( !full_path_size )
	{
		ASSERT(0);
		return false;
	}

	s.assign ( full_path );
	return true;
}

//------------------------------------------------------------------------------

} // namespace FileNamespace

//------------------------------------------------------------------------------
// FileSearchIterator
//------------------------------------------------------------------------------

FileSearchIterator::FileSearchIterator
(
	const char * search_path,
	const char * file_mask,
	bool recursive
)
:	recursive_(recursive),
	bEnd_(false),
	bFileIterator_(false),
	bDirectoryIterator_(false),
	directory_(0),
	file_iter_(0),
	bCurrentDirectory_(FALSE)
{
	if ( search_path )
		search_path_.assign(search_path);
	if ( file_mask )
		file_mask_.assign(file_mask);

	// store current directory

	getCurrentDirectory ( old_current_directory_ );

// set current directory
// if search path empty than use current directory

	if ( !search_path_.empty() )
		bCurrentDirectory_ = SetCurrentDirectory ( search_path_.c_str() );
	else
		bCurrentDirectory_ = true;

}

//------------------------------------------------------------------------------

FileSearchIterator::~FileSearchIterator()
{
	if ( !old_current_directory_.empty() )
		SetCurrentDirectory ( old_current_directory_.c_str() );
}

//------------------------------------------------------------------------------

bool FileSearchIterator::setNextSearchDirectory()
{
	const char * each_directory = "*";
	
	ASSERT ( recursive_ );

	try
	{
		WIN32_FIND_DATA findFileData;

		directory_stack_.push ( PWin32FileIteratorWithPred
			( new Win32FileIteratorWithPred
				( each_directory, FileNamespace::isSubDirectory ) ) );
		std::string current_directory;
		if ( getCurrentDirectory ( current_directory ) )
			current_directory_stack_.push ( current_directory );

next:
		if ( !directory_stack_.empty()
			&& directory_stack_.top()->next ( findFileData ) )
		{

			bCurrentDirectory_ = SetCurrentDirectory ( findFileData.cFileName );


			if ( !bCurrentDirectory_ )
				goto next;

			return true;
		}
		else if ( !directory_stack_.empty() )
		{
			directory_stack_.pop();
			current_directory_stack_.pop();
			
			if ( !current_directory_stack_.empty() )
			{
				bCurrentDirectory_ = SetCurrentDirectory
					( current_directory_stack_.top().c_str() );

	
				goto next;
			}
		}
	}
	catch ( std::exception & )
	{}
	catch ( ... )
	{}

	return false;
}

//------------------------------------------------------------------------------

// virtual
PFile FileSearchIterator::next()
{
	try
	{
		if ( bEnd_ )
			return 0;

		if ( !bCurrentDirectory_ )
			return 0;

next:
		if ( !bFileIterator_ )
		{
			file_iter_ = PWin32FileIteratorWithPred
				( new Win32FileIteratorWithPred
					( file_mask_.c_str(), FileNamespace::isFile ) );
			bFileIterator_ = true;
		}

		WIN32_FIND_DATA findFileData;

		if ( file_iter_ && file_iter_->next ( findFileData ) )
			return PFile ( new File ( findFileData.cFileName ) );

		if ( recursive_ && setNextSearchDirectory() )
		{
			bFileIterator_ = false;
			goto next;
		}
	}		
	catch ( std::exception & )
	{}
	catch ( ... )
	{}
	
	bEnd_ = true;
	return 0;
}

//------------------------------------------------------------------------------
// DirectoryIterator
//------------------------------------------------------------------------------

DirectoryIterator::DirectoryIterator
(
	const char * search_path
)
: bEnd_(false)
,	directory_iter_(0)
,	bCurrentDirectory_(FALSE)
{
	if ( search_path )
		search_path_.assign(search_path);

	// store current directory

	getCurrentDirectory ( old_current_directory_ );

// set current directory
// if search path empty than use current directory

	if ( !search_path_.empty() )
		bCurrentDirectory_ = SetCurrentDirectory ( search_path_.c_str() );
	else
		bCurrentDirectory_ = true;

}

//------------------------------------------------------------------------------

DirectoryIterator::~DirectoryIterator()
{
	if ( !old_current_directory_.empty() )
		SetCurrentDirectory ( old_current_directory_.c_str() );
}

//------------------------------------------------------------------------------

// virtual
PFile DirectoryIterator::next()
{
	const char * each_directory = "*";

	try
	{
		if ( bEnd_ )
			return 0;

		if ( !bCurrentDirectory_ )
			return 0;

		if ( !directory_iter_ )
		{
			directory_iter_ = PWin32FileIteratorWithPred
				( new Win32FileIteratorWithPred
					( each_directory, FileNamespace::isSubDirectory ) );
		}

		WIN32_FIND_DATA findFileData;

		if ( directory_iter_ && directory_iter_->next ( findFileData ) )
			return PFile ( new File ( findFileData.cFileName ) );
	}	
	catch ( std::exception & )
	{}
	catch ( ... )
	{}
	
	bEnd_ = true;
	return 0;
}

//------------------------------------------------------------------------------
// FileManager
//------------------------------------------------------------------------------

FileManager::FileManager()
{}

//------------------------------------------------------------------------------

FileManager::~FileManager()
{}

//------------------------------------------------------------------------------

// singleton support

FileManager& FileManager::getInstance()
{
	static FileManager file_man;
	return file_man;
}

//------------------------------------------------------------------------------

PFile FileManager::getFile ( const char * file_path )
{
	PFile file = 0;
	try
	{
		file = PFile ( new File ( file_path ) );
	}
	catch ( std::exception & )
	{}
	catch ( ... )
	{}

	return file;
}

//------------------------------------------------------------------------------

PFile FileManager::getFile
	( const char * directory, const char * relative_path )
{
	std::string file_path = directory;
	AppendSlash ( file_path );
	file_path += relative_path;
	return getFile ( file_path.c_str() );
}

//------------------------------------------------------------------------------

PFileIterator FileManager::createFileIterator
	( const char * file_path, const char * file, bool recursive )
{
	return PFileIterator
		( new FileSearchIterator ( file_path, file, recursive ) );
}

//------------------------------------------------------------------------------

// static
bool FileManager::getCurrentDirectory ( std::string & current_directory )
{
	return FileNamespace::getCurrentDirectory ( current_directory );
}

//------------------------------------------------------------------------------

// static
bool FileManager::setCurrentDirectory ( const std::string & current_directory )
{
	return ::SetCurrentDirectory ( current_directory.c_str() ) != 0;
}

//------------------------------------------------------------------------------

// static
bool FileManager::normalizePath ( std::string & path )
{
	bool bRet = false;
	std::string directory;
	if ( getCurrentDirectory ( directory ) )
	{
		if ( setCurrentDirectory ( path.c_str() ) )
		{
			bRet = getCurrentDirectory ( path );
		}
		setCurrentDirectory ( directory.c_str() );
	}
	return bRet;
}

//------------------------------------------------------------------------------

// static
bool FileManager::isDirectoryExist ( const char * path )
{
	DWORD dword = GetFileAttributes ( path );
	return ( dword != INVALID_FILE_ATTRIBUTES
		&& ( ( dword & FILE_ATTRIBUTE_DIRECTORY ) != 0 ) );
}

//------------------------------------------------------------------------------

// static
bool FileManager::createDirectory ( const char * path )
{
	return ::CreateDirectory ( path, NULL ) != 0;
}

//------------------------------------------------------------------------------

// static
bool FileManager::createDirectoryRecursive ( const char * path )
{
	if ( isDirectoryExist ( path ) )
		return false;
	std::string directory;
	if ( GetDirectoryFromPath ( path, directory ) )
	{
		if ( !isDirectoryExist ( directory.c_str() ) )
		{
			if ( !createDirectoryRecursive ( directory.c_str() ) )
				return false;
		}
		return createDirectory ( path );
	}
	return false;
}

//------------------------------------------------------------------------------
